@@ -25,9 +25,11 @@ import distro_info
 import glob
 import logging
 import os
+import pwd
 import re
 import hashlib
 import subprocess
+import pathlib
 from subprocess import PIPE, Popen
 
 from .utils import get_arch
@@ -55,6 +57,7 @@ class DistUpgradeQuirks(object):
         self._snap_list = None
         self._from_version = None
         self._to_version = None
+        self._did_change_font = False
 
     # the quirk function have the name:
     #  $Name (e.g. PostUpgrade)
@@ -111,21 +114,27 @@ class DistUpgradeQuirks(object):
         logging.debug("running Quirks.PreCacheOpen")
 
     # individual quirks handler that run *after* the cache is opened
-    def hirsutePostInitialUpdate(self):
+    def noblePostInitialUpdate(self):
         # PreCacheOpen would be better but controller.abort fails terribly
         """ run after the apt cache is opened the first time """
-        logging.debug("running Quirks.hirsutePostInitialUpdate")
+        logging.debug("running Quirks.noblePostInitialUpdate")
         self._get_from_and_to_version()
         self._test_and_fail_on_i386()
         self._test_and_fail_on_aufs()
+        self._test_and_fail_on_power8()
+        self._test_and_fail_on_bios_with_xfs_boot()
 
         cache = self.controller.cache
         self._test_and_warn_if_ros_installed(cache)
 
+        self._maybe_prevent_flatpak_auto_removal()
+
         if 'snapd' not in cache:
             logging.debug("package required for Quirk not in cache")
             return
-        if cache['snapd'].is_installed:
+        if cache['snapd'].is_installed and \
+                (os.path.exists('/run/snapd.socket') or
+                 os.path.exists('/run/snapd-snap.socket')):
             self._checkStoreConnectivity()
         # If the snap store is accessible, at the same time calculate the
         # extra size needed by to-be-installed snaps.  This also prepares
@@ -133,8 +142,8 @@ class DistUpgradeQuirks(object):
         if self._snapstore_reachable:
             self._calculateSnapSizeRequirements()
 
-    def hirsutePostUpgrade(self):
-        logging.debug("running Quirks.hirsutePostUpgrade")
+    def noblePostUpgrade(self):
+        logging.debug("running Quirks.noblePostUpgrade")
         cache = self.controller.cache
         if 'snapd' not in cache:
             logging.debug("package required for Quirk not in cache")
@@ -145,6 +154,9 @@ class DistUpgradeQuirks(object):
         if 'ubuntu-desktop-raspi' in cache:
             if cache['ubuntu-desktop-raspi'].is_installed:
                 self._replace_fkms_overlay()
+        if 'linux-firmware-raspi2' in cache:
+            if cache['linux-firmware-raspi2'].is_installed:
+                self._remove_uboot_on_rpi()
 
     # individual quirks handler when the dpkg run is finished ---------
     def PostCleanup(self):
@@ -158,16 +170,16 @@ class DistUpgradeQuirks(object):
         self._removeOldApportCrashes()
         self._killUpdateNotifier()
         self._killKBluetooth()
-        self._killScreensaver()
         self._pokeScreensaver()
         self._stopDocvertConverter()
+        self._set_generic_font()
 
     # individual quirks handler that run *right before* the dist-upgrade
     # is calculated in the cache
     def PreDistUpgradeCache(self):
         """ run right before calculating the dist-upgrade """
         logging.debug("running Quirks.PreDistUpgradeCache")
-        self._install_python_is_python2()
+        # self._install_python_is_python2()
 
     # individual quirks handler that run *after* the dist-upgrade was
     # calculated in the cache
@@ -366,6 +378,35 @@ class DistUpgradeQuirks(object):
                 self._view.error(summary, msg)
                 self.controller.abort()
 
+    def _test_and_fail_on_power8(self):
+        """
+        Test and fail if running on a POWER8 processor. Starting with Ubuntu
+        22.04, ppc64el is compiled for POWER9 processors. Prior to 22.04 it
+        was/is compiled for POWER8 processors. Hence POWER8 processors are not
+        able to run, install or upgrade to jammy.
+        """
+        if self.arch != "ppc64el":
+            return
+
+        updates_end = ""
+        if self._from_version == "20.04":
+            updates_end = "April 2025"
+        elif self._from_version == "21.10":
+            updates_end = "July 2022"
+
+        with open("/proc/cpuinfo") as f:
+            cpuinfo = f.read()
+
+            if re.search("^cpu\s*:\s*POWER8", cpuinfo, re.MULTILINE):
+                logging.error("POWER8 processor detected")
+                summary = _("Sorry, no more upgrades for this system")
+                msg = _("There will not be any further Ubuntu releases "
+                        "for this system's POWER8 processor.\n\n"
+                        "Updates for Ubuntu %s will continue until %s." %
+                        (self._from_version, updates_end))
+                self._view.error(summary, msg)
+                self.controller.abort()
+
     def _test_and_warn_if_vserver(self):
         """
         upstart and vserver environments are not a good match, warn
@@ -435,6 +476,74 @@ class DistUpgradeQuirks(object):
             if not res:
                 self.controller.abort()
 
+    def _maybe_prevent_flatpak_auto_removal(self):
+        """
+        If flatpak is installed, and there are either active remotes, or
+        flatpak apps installed, prevent flatpak's auto-removal on upgrade.
+        """
+        prevent_auto_removal = False
+
+        if "flatpak" not in self.controller.cache:
+            return
+
+        if not self.controller.cache["flatpak"].is_installed:
+            return
+
+        if not os.path.exists("/usr/bin/flatpak"):
+            return
+
+        for subcmd in ["remotes", "list"]:
+            r = subprocess.run(
+                    ["/usr/bin/flatpak", subcmd],
+                    stdout=subprocess.PIPE
+            )
+            if r.stdout.decode("utf-8").strip():
+                prevent_auto_removal = True
+                break
+
+        logging.debug("flatpak will{}be marked as manually installed"
+                      .format(" " if prevent_auto_removal else " NOT "))
+
+        if not prevent_auto_removal:
+            return
+
+        self.controller.cache["flatpak"].mark_auto(auto=False)
+
+        for pkg in ("plasma-discover-backend-flatpak",
+                    "gnome-software-plugin-flatpak"):
+            if pkg not in self.controller.cache:
+                continue
+
+            if not self.controller.cache[pkg].is_installed:
+                continue
+
+            logging.debug("{} will be marked as manually installed"
+                          .format(pkg))
+            self.controller.cache[pkg].mark_auto(auto=False)
+
+        self.controller.cache.commit(
+            self._view.getAcquireProgress(),
+            self._view.getInstallProgress(self.controller.cache)
+        )
+
+    def _add_apport_ignore_tracker_extract(self):
+        path = '/etc/apport/blacklist.d/upgrade-quirk-tracker-extract-3'
+
+        try:
+            os.makedirs(os.path.dirname(path), exist_ok=True)
+            with open(path, 'w') as f:
+                f.write('/usr/libexec/tracker-extract-3\n')
+        except Exception as e:
+            logging.debug('Failed to create {}: {}'.format(path, e))
+
+    def _remove_apport_ignore_tracker_extract(self):
+        path = '/etc/apport/blacklist.d/upgrade-quirk-tracker-extract-3'
+
+        try:
+            os.remove(path)
+        except Exception as e:
+            logging.debug('Failed to remove {}: {}'.format(path, e))
+
     def _checkArmCPU(self):
         """
         parse /proc/cpuinfo and search for ARMv6 or greater
@@ -450,6 +559,19 @@ class DistUpgradeQuirks(object):
             return False
         return True
 
+    def _readDMIVendor(self):
+        """
+        read /sys/class/dmi/id/sys_vendor in case there are
+        system specific issues with upgrading e.g LP: #1928434
+        """
+        logging.debug("checking for system vendor")
+        if not os.path.exists("/sys/class/dmi/id/sys_vendor"):
+            logging.error("cannot open /sys/class/dmi/id/sys_vendor")
+            return ''
+        with open("/sys/class/dmi/id/sys_vendor") as f:
+            vendor = f.read()
+        return vendor
+
     def _stopApparmor(self):
         """ /etc/init.d/apparmor stop (see bug #559433)"""
         if os.path.exists("/etc/init.d/apparmor"):
@@ -475,12 +597,6 @@ class DistUpgradeQuirks(object):
             logging.debug("killing kblueplugd kbluetooth4")
             subprocess.call(["killall", "-q", "kblueplugd", "kbluetooth4"])
 
-    def _killScreensaver(self):
-        """killall gnome-screensaver """
-        if os.path.exists("/usr/bin/killall"):
-            logging.debug("killing gnome-screensaver")
-            subprocess.call(["killall", "-q", "gnome-screensaver"])
-
     def _pokeScreensaver(self):
         if (os.path.exists("/usr/bin/xdg-screensaver") and
                 os.environ.get('DISPLAY')):
@@ -574,6 +690,8 @@ class DistUpgradeQuirks(object):
         elif 'error: access denied' in connected[1]:
             res = False
             logging.error("Not running as root!")
+        else:
+            logging.error("Unhandled error connecting to the snap store.")
         if not res:
             self.controller.abort()
 
@@ -624,7 +742,9 @@ class DistUpgradeQuirks(object):
         # _calculateSnapSizeRequirements call.
         for snap, snap_object in self._snap_list.items():
             command = snap_object['command']
-            if command == 'refresh':
+            if command == 'switch':
+                # TODO: This status should be updated, but the time of
+                # this change to snap switch is post-translation freeze.
                 self._view.updateStatus(_("refreshing snap %s" % snap))
                 popenargs = ["snap", command,
                              "--channel", snap_object['channel'], snap]
@@ -955,8 +1075,34 @@ class DistUpgradeQuirks(object):
         for old, new in replacements:
             if old in cache and cache[old].is_installed:
                 if new:
-                    solver.clear(cache[new])
-                    solver.protect(cache[new])
+                    if new not in cache:
+                        old_inst = [old for old, new in replacements if old
+                                    in cache and cache[old].is_installed]
+                        summary = _("universe component not enabled")
+                        msg = _("You have the package %s installed which "
+                                "is a python2 package. python2 has been "
+                                "deprecated in Ubuntu 20.04 LTS and is now "
+                                "available from the universe component of "
+                                "the Ubuntu archive. To ensure any existing "
+                                "third party code you have installed "
+                                "continues to work the release upgrade "
+                                "process would install %s for you. However, "
+                                "the universe component of the archive is not "
+                                "enabled on this system. For the upgrade "
+                                "process to proceed you should either enable "
+                                "the universe component in "
+                                "/etc/apt/sources.list or remove %s"
+                                % (old, new, ', '.join(old_inst))
+                                )
+                        logging.info(summary)
+                        # necessary to make the error text available again on
+                        # stdout for the text frontend
+                        cache._stopAptResolverLog()
+                        self._view.error(summary, msg)
+                        self.controller.abort()
+                    else:
+                        solver.clear(cache[new])
+                        solver.protect(cache[new])
                 solver.clear(cache[old])
                 solver.remove(cache[old])
 
@@ -973,6 +1119,25 @@ class DistUpgradeQuirks(object):
                 logging.warning(msg)
                 apt.apt_pkg.config.set("Apt::Install-Recommends", "1")
 
+    def _is_deb2snap_metapkg_installed(self, deb2snap_entry):
+        """ Helper function that checks if the given deb2snap entry
+            has at least one metapkg which is installed on the system.
+        """
+        metapkg_list = deb2snap_entry.get("metapkg", None)
+        if not isinstance(metapkg_list, list):
+            metapkg_list = [metapkg_list]
+
+        for metapkg in metapkg_list:
+            if metapkg not in self.controller.cache:
+                continue
+            if metapkg and \
+                    self.controller.cache[metapkg].is_installed is False:
+                continue
+
+            return True
+
+        return False
+
     def _prepare_snap_replacement_data(self):
         """ Helper function fetching all required info for the deb-to-snap
             migration: version strings for upgrade (from and to) and the list
@@ -993,11 +1158,7 @@ class DistUpgradeQuirks(object):
 
             for snap in d2s["seeded"]:
                 seed = d2s["seeded"][snap]
-                metapkg = seed.get("metapkg", None)
-                if metapkg not in self.controller.cache:
-                    continue
-                if metapkg and \
-                        self.controller.cache[metapkg].is_installed is False:
+                if not self._is_deb2snap_metapkg_installed(seed):
                     continue
                 deb = seed.get("deb", None)
                 from_chan = seed.get("from_channel", from_channel)
@@ -1007,11 +1168,7 @@ class DistUpgradeQuirks(object):
             for snap in d2s["unseeded"]:
                 unseed = d2s["unseeded"][snap]
                 deb = unseed.get("deb", None)
-                metapkg = unseed.get("metapkg", None)
-                if metapkg not in self.controller.cache:
-                    continue
-                if metapkg and \
-                        self.controller.cache[metapkg].is_installed is False:
+                if not self._is_deb2snap_metapkg_installed(unseed):
                     continue
                 from_chan = unseed.get("from_channel", from_channel)
                 unseeded_snaps[snap] = (deb, from_chan)
@@ -1044,13 +1201,13 @@ class DistUpgradeQuirks(object):
             self._view.processEvents()
             if re.search("^installed: ", snap_info[0], re.MULTILINE):
                 logging.debug("Snap %s is installed" % snap)
-                # its not tracking the release channel so don't refresh
+                # its not tracking the release channel so don't switch
                 if not re.search(r"^tracking:.*%s" % from_channel,
                                  snap_info[0], re.MULTILINE):
                     logging.debug("Snap %s is not tracking the release channel"
                                   % snap)
                     continue
-                snap_object['command'] = 'refresh'
+                snap_object['command'] = 'switch'
             else:
                 # Do not replace packages not installed
                 cache = self.controller.cache
@@ -1107,7 +1264,7 @@ class DistUpgradeQuirks(object):
                             logging.debug("Snap %s is being used by %s. "
                                           "Switching it to stable track"
                                           % (snap, plug_snap))
-                            snap_object['command'] = 'refresh'
+                            snap_object['command'] = 'switch'
                             snap_object['channel'] = 'stable'
                             break
 
@@ -1116,7 +1273,7 @@ class DistUpgradeQuirks(object):
 
     def _replace_fkms_overlay(self, boot_dir='/boot/firmware'):
         failure_action = (
-            "you may need to replace the vc4-fkms-v3d overlay with "
+            "You may need to replace the vc4-fkms-v3d overlay with "
             "vc4-kms-v3d in config.txt on your boot partition")
 
         try:
@@ -1164,3 +1321,326 @@ class DistUpgradeQuirks(object):
         except IOError as exc:
             logging.error("unable to write new boot config to %s: %s; %s",
                           boot_config_filename, exc, failure_action)
+
+    def _remove_uboot_on_rpi(self, boot_dir='/boot/firmware'):
+        kernel_line = 'kernel=vmlinuz'
+        initramfs_line = 'initramfs initrd.img followkernel'
+        failure_action = (
+            "You may need to replace u_boot_* with vmlinuz, and add "
+            "{initramfs_line!r} to config.txt on your boot "
+            "partition; see LP: #1936401 for further details")
+        failure_action = failure_action.format(initramfs_line=initramfs_line)
+        change_prefix = '# commented by do-release-upgrade (LP: #1936401)'
+        added_prefix = '# added by do-release-upgrade (LP: #1936401)'
+        merge_prefix = '# merged from {} by do-release-upgrade (LP: #1936401)'
+
+        try:
+            boot_config_filename = os.path.join(boot_dir, 'config.txt')
+            with open(boot_config_filename, 'r', encoding='utf-8') as f:
+                boot_config = f.read()
+        except FileNotFoundError:
+            logging.error("failed to open boot configuration in %s; %s",
+                          boot_config_filename, failure_action)
+            return
+
+        def replace_uboot(lines):
+            result = []
+            removed_uboot = added_kernel = False
+            for line in lines:
+                if line == '[all]':
+                    # Add the vmlinuz kernel and initrd.img to the first
+                    # encountered [all] section, just in case the user has any
+                    # custom kernel overrides later in the sequence
+                    result.append(line)
+                    if not added_kernel:
+                        result.append(added_prefix)
+                        result.append(kernel_line)
+                        result.append(initramfs_line)
+                        added_kernel = True
+                elif line.startswith('device_tree_address='):
+                    # Disable any device_tree_address= line (leave the boot
+                    # firmware to place this rather than risk trampling over it
+                    # with kernel/initrd)
+                    result.append(change_prefix)
+                    result.append('#' + line)
+                elif line.startswith('kernel=uboot_rpi_'):
+                    # Disable all kernel=uboot_rpi_* lines found in the
+                    # configuration
+                    removed_uboot = True
+                    result.append(change_prefix)
+                    result.append('#' + line)
+                else:
+                    result.append(line)
+            # We don't *want* to touch the config unless we absolutely have to,
+            # and the user may have already performed this conversion (e.g. to
+            # take advantage of USB MSD boot), or otherwise customized their
+            # boot configuration. Hence, if we didn't find (and remove) any
+            # u-boot entries, assume the kernel/initrd config was already fine
+            # and return the verbatim config
+            if removed_uboot:
+                if not added_kernel:
+                    result.append(added_prefix)
+                    result.append('[all]')
+                    result.append(kernel_line)
+                    result.append(initramfs_line)
+                return result
+            else:
+                return lines
+
+        def merge_includes(lines):
+            result = []
+            skip_comments = True
+            found_includes = False
+            for line in lines:
+                # Skip the initial comment block warning that config.txt is not
+                # to be edited (the usercfg.txt and syscfg.txt files were
+                # merged in the groovy release, but not for upgraders)
+                if line.startswith('#') and skip_comments:
+                    continue
+                skip_comments = False
+                if line in ('include syscfg.txt', 'include usercfg.txt'):
+                    # Merge the usercfg.txt and syscfg.txt configuration files
+                    # into config.txt, skipping the initial comments. Note that
+                    # we don't bother with other includes the user has added
+                    # themselves (as they presumably want to keep those)
+                    found_includes = True
+                    included_filename = line.split(maxsplit=1)[1]
+                    result.append(merge_prefix.format(included_filename))
+                    included_filename = os.path.join(boot_dir,
+                                                     included_filename)
+                    skip_comments = True
+                    with open(included_filename, 'r', encoding='utf-8') as f:
+                        for line in f:
+                            if line.startswith('#') and skip_comments:
+                                continue
+                            skip_comments = False
+                            result.append(line.rstrip())
+                    target_filename = included_filename + '.distUpgrade'
+                    try:
+                        os.rename(included_filename, target_filename)
+                    except IOError as exc:
+                        logging.error("failed to move included configuration "
+                                      "from %s to %s; %s", included_filename,
+                                      target_filename, exc)
+                else:
+                    result.append(line)
+            # Again, if we found no includes, return the original verbatim
+            # (i.e. with initial comments intact)
+            if found_includes:
+                return result
+            else:
+                return lines
+
+        lines = [line.rstrip() for line in boot_config.splitlines()]
+        lines = replace_uboot(merge_includes(lines))
+        new_config = ''.join(line + '\n' for line in lines)
+
+        if new_config == boot_config:
+            logging.warning("no u-boot removal performed in %s",
+                            boot_config_filename)
+            return
+
+        try:
+            boot_backup_filename = boot_config_filename + '.distUpgrade'
+            with open(boot_backup_filename, 'w', encoding='utf-8') as f:
+                f.write(boot_config)
+        except IOError as exc:
+            logging.error("unable to write boot config backup to %s: %s; %s",
+                          boot_backup_filename, exc, failure_action)
+            return
+        try:
+            with open(boot_config_filename, 'w', encoding='utf-8') as f:
+                f.write(new_config)
+        except IOError as exc:
+            logging.error("unable to write new boot config to %s: %s; %s",
+                          boot_config_filename, exc, failure_action)
+
+    def _set_generic_font(self):
+        """ Due to changes to the Ubuntu font we enable a generic font
+            (in practice DejaVu or Noto) during the upgrade.
+            See https://launchpad.net/bugs/2034986
+        """
+        temp_font = 'Sans'
+
+        if self._did_change_font:
+            return
+
+        if os.getenv('XDG_SESSION_TYPE', '') in ('', 'tty'):
+            # Avoid running this on server systems or when the upgrade
+            # is done over ssh.
+            return
+
+        if not {'SUDO_UID', 'PKEXEC_UID'}.intersection(os.environ):
+            logging.debug(
+                'Cannot determine how root privileges were gained, will '
+                'not change font'
+            )
+            return
+        try:
+            uid = int(os.getenv('SUDO_UID', os.getenv('PKEXEC_UID')))
+            pwuid = pwd.getpwuid(uid)
+        except ValueError:
+            logging.debug(
+                'Cannot determine non-root UID, will not change font'
+            )
+            return
+
+        schema = 'org.gnome.desktop.interface'
+
+        desktops = os.getenv('XDG_CURRENT_DESKTOP', '').split(':')
+
+        # Some flavors use other schemas for the desktop font.
+        if 'MATE' in desktops or 'UKUI' in desktops:
+            schema = 'org.mate.interface'
+        elif 'X-Cinnamon' in desktops:
+            schema = 'org.cinnamon.desktop.interface'
+
+        # Some flavors lack the systemd integration needed for a
+        # user service, so we create an autostart file instead.
+        use_autostart = bool(
+            set(['Budgie', 'LXQt', 'MATE', 'UKUI', 'X-Cinnamon', 'XFCE'])
+            & set(desktops)
+        )
+
+        r = subprocess.run(
+            ['systemd-run', '--user', '-M', f'{pwuid.pw_name}@.host',
+             '--wait', '--pipe', '-q', '--',
+             '/usr/bin/gsettings', 'get',
+             f'{schema}', 'font-name'],
+            stdout=subprocess.PIPE,
+            encoding='utf-8',
+        )
+
+        (font, _, size) = r.stdout.strip('\'\n').rpartition(' ')
+        font = font or 'Ubuntu'
+        try:
+            int(size)
+        except ValueError:
+            size = '11'
+
+        logging.debug(f'Setting generic font {temp_font} {size} during the '
+                      f'upgrade. Original font is {font} {size}.')
+
+        r = subprocess.run([
+            'systemd-run', '--user', '-M', f'{pwuid.pw_name}@.host',
+            '--wait', '--pipe', '-q', '--',
+            '/usr/bin/gsettings', 'set', f'{schema}',
+            'font-name', f'"{temp_font} {size}"'
+        ])
+        if r.returncode != 0:
+            logging.debug(f'Failed to change font to {temp_font} {size}')
+            return
+
+        self._did_change_font = True
+
+        # Touch a file to indiate that the font should be restored on the next
+        # boot.
+        need_font_restore_file = os.path.join(
+            pwuid.pw_dir, '.config/upgrade-need-font-restore'
+        )
+        os.makedirs(os.path.dirname(need_font_restore_file), exist_ok=True)
+        pathlib.Path(need_font_restore_file).touch(mode=0o666)
+        os.chown(need_font_restore_file, pwuid.pw_uid, pwuid.pw_gid)
+
+        if use_autostart:
+            autostart_file = '/etc/xdg/autostart/upgrade-restore-font.desktop'
+            os.makedirs(os.path.dirname(autostart_file), exist_ok=True)
+            flag = '$HOME/.config/upgrade-need-font-restore'
+            with open(autostart_file, 'w') as f:
+                f.write(
+                    '[Desktop Entry]\n'
+                    'Name=Restore font after upgrade\n'
+                    'Comment=Auto-generated by ubuntu-release-upgrader\n'
+                    'Type=Application\n'
+                    f'Exec=sh -c \'if [ -e "{flag}" ]; then gsettings set '
+                    f'{schema} font-name "{font} {size}";'
+                    f'rm -f "{flag}"; fi\'\n'
+                    'NoDisplay=true\n'
+                )
+            return
+
+        # If we set the font back to normal before a reboot, the font will
+        # still get all messed up. To allow normal usage whether the user
+        # reboots immediately or not, create a service that will run only if a
+        # ~/.config/upgrade-need-font-restore exists, and then remove that file
+        # in ExecStart. This has the effect of creating a one-time service on
+        # the next boot.
+        unit_file = '/usr/lib/systemd/user/upgrade-restore-font.service'
+        os.makedirs(os.path.dirname(unit_file), exist_ok=True)
+
+        with open(unit_file, 'w') as f:
+            f.write(
+                '# Auto-generated by ubuntu-release-upgrader\n'
+                '[Unit]\n'
+                'Description=Restore font after upgrade\n'
+                'After=graphical-session.target dconf.service\n'
+                'ConditionPathExists=%h/.config/upgrade-need-font-restore\n'
+                '\n'
+                '[Service]\n'
+                'Type=oneshot\n'
+                'ExecStart=/usr/bin/gsettings set '
+                f'{schema} font-name \'{font} {size}\'\n'
+                'ExecStart=/usr/bin/rm -f '
+                '%h/.config/upgrade-need-font-restore\n'
+                '\n'
+                '[Install]\n'
+                'WantedBy=graphical-session.target\n'
+            )
+
+        subprocess.run([
+            'systemctl',
+            '--user', '-M', f'{pwuid.pw_name}@.host',
+            'daemon-reload'
+        ])
+        r = subprocess.run([
+            'systemctl',
+            '--user', '-M', f'{pwuid.pw_name}@.host',
+            'enable', os.path.basename(unit_file)
+        ])
+        if r.returncode != 0:
+            logging.debug(f'Failed to enable {os.path.basename(unit_file)}. '
+                          'Font will not be restored on reboot')
+
+    def _test_and_fail_on_bios_with_xfs_boot(self):
+        """
+        If this system is BIOS and has XFS on /boot, do not allow the upgrade
+        to happen (LP: #2039172).
+        """
+        if os.path.isdir('/sys/firmware/efi'):
+            # UEFI boot, nothing to do.
+            return
+
+        boot_fs = None
+        root_fs = None
+        with open('/proc/mounts') as f:
+            for line in f.readlines():
+                try:
+                    (what, where, fs, opts, a, b) = line.strip().split()
+                    where = where.lower()
+                    fs = fs.lower()
+                except ValueError as e:
+                    logging.debug(
+                        f'line \'{line}\' in /proc/mounts not understood'
+                    )
+                    continue
+
+                if where == '/':
+                    root_fs = fs
+                elif where == '/boot':
+                    boot_fs = fs
+
+                if boot_fs and root_fs:
+                    break
+
+        if boot_fs == 'xfs' or (boot_fs is None and root_fs == 'xfs'):
+            # If we got to this point, then this is a BIOS system, and /boot is
+            # XFS or / is XFS and there is no separate /boot partition. Do not
+            # allow the upgrade to happen.
+            logging.error('Cannot upgrade system with BIOS and XFS /boot')
+            self._view.error(
+                _("Cannot upgrade this system"),
+                _("Due to a bug in grub, this system cannot "
+                  "be safely upgraded at this time.\n\n"
+                  "See https://launchpad.net/bugs/2039172.")
+            )
+            self.controller.abort()
